#!/usr/bin/env python
"""
Synchronise a project TEMPLATE branch with the template.
"""
import fnmatch
import logging
import sys
from configparser import ConfigParser, NoSectionError
from distutils.dir_util import copy_tree
from subprocess import Popen, PIPE
from typing import Tuple
import git
import os
import shutil
import tempfile
from pathlib import Path
from github import Github, GithubException
from packaging import version
from rich import print

from mlf_core.config.config import ConfigCommand
from mlf_core.create.github_support import decrypt_pat, load_github_username, create_sync_secret
from mlf_core.common.load_yaml import load_yaml_file
from mlf_core.create.create import choose_domain
from mlf_core.common.version import load_project_template_version_and_handle, load_mlf_core_template_version
from mlf_core.custom_cli.questionary import mlf_core_questionary_or_dot_mlf_core

log = logging.getLogger(__name__)


class TemplateSync:
    """
    Hold syncing information and results.

    project_dir (str): The path to the mlf-core project root directory
    from_branch (str): Original branch
    project_dir (str): Path to target project directory
    from_branch (str): Original branch
    original_branch (str): Repo branch that was checked out before we started.
    made_changes (bool): Whether making the new template project introduced any changes
    gh_username (str): GitHub username
    patch_update (bool): Whether a patch update was found for the template or not
    minor_update (bool): Whether a minor update was found for the template or not
    major_update (bool): Whether a major update was found for the template or not
    repo_owner (str): Owner of the repo (either orga name or personal github username)
    """

    def __init__(self,
                 project_dir,
                 new_template_version,
                 from_branch=None,
                 gh_username=None,
                 token=None,
                 major_update=False,
                 minor_update=False,
                 patch_update=False):
        self.project_dir = os.path.abspath(project_dir)
        self.from_branch = from_branch
        self.original_branch = None
        self.made_changes = False
        self.gh_pr_returned_data = {}
        self.major_update = major_update
        self.minor_update = minor_update
        self.patch_update = patch_update
        self.gh_username = gh_username if gh_username else load_github_username()
        self.token = token if token else decrypt_pat()
        self.dot_mlf_core = {}
        self.repo_owner = self.gh_username
        self.new_template_version = new_template_version
        self.github = Github(self.token)
        self.blacklisted_globs = []

    def sync(self):
        """
        Sync the mlf-core project
        """
        self.inspect_sync_dir()
        # get blacklisted files on current working branch
        self.blacklisted_globs = self.get_blacklisted_sync_globs()
        self.checkout_template_branch()
        self.delete_template_branch_files()
        self.make_template_project()
        self.commit_template_changes()

        # Push and make a pull request
        if self.made_changes:
            try:
                self.push_template_branch()
                self.make_pull_request()
            except Exception as e:
                self.reset_target_dir()
                print(f'[bold red]{e}')
                sys.exit(1)

        self.reset_target_dir()

        if not self.made_changes:
            print('[bold blue]No changes made to TEMPLATE - sync complete')

    def inspect_sync_dir(self):
        """
        Examines target directory to sync, verifies that it is a git repository and ensures that there are no uncommitted changes.
        """
        if not os.path.exists(os.path.join(str(self.project_dir), '.mlf_core.yml')):
            print(f'[bold red]{self.project_dir} does not appear to contain a .mlf_core.yml file. Did you delete it?')
            sys.exit(1)
        # store .mlf_core.yml content for later reuse in the dry create run
        self.dot_mlf_core = load_yaml_file(os.path.join(str(self.project_dir), '.mlf_core.yml'))
        log.debug(f'Loaded .mlf_core.yml file content. Content is: {self.dot_mlf_core}')
        # Check that the project_dir is a git repo
        try:
            self.repo = git.Repo(self.project_dir)
        except git.exc.InvalidGitRepositoryError:
            print(f'[bold red]{self.project_dir} does not appear to be a git repository.')
            sys.exit(1)

        # get current branch so we can switch back later
        self.original_branch = self.repo.active_branch.name
        print(f'[bold blue]Original Project repository branch is {self.original_branch}')

        # Check to see if there are uncommitted changes on current branch
        if self.repo.is_dirty(untracked_files=True):
            print('[bold red]Uncommitted changes found in Project directory!\nPlease commit these before running mlf-core sync')
            sys.exit(1)

    def checkout_template_branch(self):
        """
        Try to check out the origin/TEMPLATE in a new TEMPLATE branch.
        If this fails, try to check out an existing local TEMPLATE branch.
        """
        try:
            self.from_branch = self.repo.active_branch.name
        except git.exc.GitCommandError as e:
            print(f'[bold red]Could not find active repo branch:\n{e}')
        # Try to check out the `TEMPLATE` branch
        try:
            self.repo.git.checkout('origin/TEMPLATE', b='TEMPLATE')
        except git.exc.GitCommandError:
            # Try to check out an existing local branch called TEMPLATE
            try:
                self.repo.git.checkout('TEMPLATE')
            except git.exc.GitCommandError:
                print('[bold red]Could not check out branch "origin/TEMPLATE" or "TEMPLATE"')
                sys.exit(1)

    def delete_template_branch_files(self):
        """
        Delete all files in the TEMPLATE branch
        """
        # Delete everything
        print('[bold blue]Deleting all files in TEMPLATE branch')
        for the_file in os.listdir(self.project_dir):
            if the_file == '.git':
                log.debug('Found .git directory. Skipping deleting it.')
                continue
            file_path = os.path.join(self.project_dir, the_file)
            try:
                if os.path.isfile(file_path):
                    log.debug(f'Deleting file {file_path}')
                    os.unlink(file_path)
                elif os.path.isdir(file_path):
                    log.debug(f'Deleting directory {file_path}')
                    shutil.rmtree(file_path)
            except Exception as e:
                print(f'[bold red]{e}')
                sys.exit(1)

    def make_template_project(self):
        """
        Delete all files and make a fresh template.
        """
        print('[bold blue]Creating a new template project.')
        # dry create run from mlf-core in tmp directory
        with tempfile.TemporaryDirectory() as tmpdirname:
            old_cwd = str(Path.cwd())
            log.debug(f'Saving current working directory {old_cwd}.')
            os.chdir(tmpdirname)
            log.debug(f'Changed directory to {tmpdirname}.')
            log.debug(f'Calling choose_domain with {self.dot_mlf_core}.')
            choose_domain(path=Path.cwd(), domain=None, dot_mlf_core=self.dot_mlf_core)
            # copy into the cleaned TEMPLATE branch's project directory
            log.debug(f'Copying created template into {self.project_dir}.')
            copy_tree(os.path.join(tmpdirname, self.dot_mlf_core['project_slug']), str(self.project_dir))
            log.debug(f'Changing directory back to {old_cwd}.')
            os.chdir(old_cwd)

    def commit_template_changes(self):
        """
        If we have any changes with the new template files, make a git commit
        """
        # Check that we have something to commit
        if not self.repo.is_dirty(untracked_files=True):
            print('[bold blue]Template contains no changes - no new commit created')
            return False
        # Commit changes
        try:
            # git add only non-blacklisted files
            print('[bold blue]Staging template.')
            # add all files to stage
            self.repo.git.add(A=True)
            # get all changed/modified files during the sync (including blacklisted ones)
            changed_files = [item.a_path for item in self.repo.index.diff('HEAD')]
            blacklisted_changed_files = set()
            for pattern in self.blacklisted_globs:
                # keep track of all staged files matching a glob from the mlf_core.cfg file
                # those files will be excluded from syncing but will still be available in every new created projects
                blacklisted_changed_files |= {file for file in fnmatch.filter(changed_files, pattern)}
            nl = '\n'
            log.debug(f'Blacklisted (unsynced) files are:{nl}{nl.join(file for file in blacklisted_changed_files)}' if blacklisted_changed_files else
                      'No blacklisted files for syncing found.')
            # each blacklisted file must be first unstaged and then get its changes discarded, so it will not be marked as modified by git
            for blacklisted_file in blacklisted_changed_files:
                log.debug(f'Unstaging {blacklisted_file}')
                Popen(['git', 'reset', blacklisted_file], stdout=PIPE, stderr=PIPE, universal_newlines=True)
                log.debug(f'Discarding changes in working directory of file {blacklisted_file}')
                Popen(['git', 'checkout', '--', blacklisted_file], stdout=PIPE, stderr=PIPE, universal_newlines=True)

            files_to_commit = [file for file in changed_files if file not in blacklisted_changed_files]
            log.debug(f'Files to commit are:{nl}{nl.join(file for file in files_to_commit)}' if files_to_commit else
                      'No files to commit found.')
            if files_to_commit:
                print('[bold blue]Committing changes of non blacklisted files.')
                Popen(['git', 'commit', '-m', 'mlf-core sync', *files_to_commit], stdout=PIPE, stderr=PIPE, universal_newlines=True)
                print('[bold blue]Stashing and saving TEMPLATE branch changes!')
                Popen(['git', 'stash'], stdout=PIPE, stderr=PIPE, universal_newlines=True)
                self.made_changes = True
                print('[bold blue]Committed changes to TEMPLATE branch')

            # if a file was added to the template, but is blacklisted, it remains untracked; so it should be removed
            for untracked_file in self.repo.untracked_files:
                log.debug(f'Removing untracked file {untracked_file}')
                Path(untracked_file).unlink()
        except Exception as e:
            print(f'[bold red]Could not commit changes to TEMPLATE:\n{e}')
            sys.exit(1)
        return True

    def push_template_branch(self):
        """
        If there are any changes to the template, push the TEMPLATE branch to the default remote
        and push to the actual sync temporary branch, where a PR is actually created from to development branch.
        """
        print(f'[bold blue]Pushing TEMPLATE branch to remote: {os.path.basename(self.project_dir)}')
        try:
            log.debug('Getting origin as remote.')
            origin = self.repo.remote('origin')
            log.debug('Setting TEMPLATE branch as upstream tracking branch.')
            self.repo.head.ref.set_tracking_branch(origin.refs.TEMPLATE)
            log.debug('Pushing to upstream branch TEMPLATE.')
            self.repo.git.push(force=True)
            print(f'[bold blue]Checking out to new branch mlf_core_sync_v{self.new_template_version}')
            log.debug(f'git checkout -b mlf_core_sync_v{self.new_template_version}')
            self.repo.git.checkout('-b', f'mlf_core_sync_v{self.new_template_version}')
            log.debug(f'git push origin mlf_core_sync_v{self.new_template_version}')
            print(f'[bold blue]Pushing to remote branch mlf_core_sync_v{self.new_template_version}')
            self.repo.remotes.origin.push(refspec=f'mlf_core_sync_v{self.new_template_version}:mlf_core_sync_v{self.new_template_version}')
        except git.exc.GitCommandError as e:
            print(f'Could not push TEMPLATE or mlf_core_sync_v{self.new_template_version} branch:\n{e}')
            sys.exit(1)

    def make_pull_request(self):
        """
        Create a pull request to a base branch from a head branch (that is, a temporary branch only created for syncing the new template version)
        """
        log.debug('Preparing PR contents to submit a sync PR.')
        if self.dot_mlf_core['is_github_orga']:
            self.repo_owner = self.dot_mlf_core['github_orga']
        pr_title = f'Important mlf-core template update {self.new_template_version} released!'
        pr_body_text = (
            'A new release of the main template in mlf-core has just been released. '
            'This automated pull-request attempts to apply the relevant updates to this Project.\n\n'
            'Please make sure to merge this pull-request as soon as possible. '
            'Once complete, make a new minor release of your Project.\n\n'
            'For more information on the actual changes, read the latest mlf-core changelog.')
        log.debug(f'PR title is:\n{pr_title}')
        log.debug(f'PR body is:\n{pr_body_text}')

        # Check, whether a mlf-core sync PR is already open
        self.check_pull_request_exists()
        # Submit the new pull request with the latest mlf-core sync changes
        self.submit_pull_request(pr_title, pr_body_text)

    def submit_pull_request(self, pr_title: str, pr_body_text: str):
        """
        Create a new pull-request on GitHub
        """
        repo = self.github.get_repo(f'{self.repo_owner}/{self.dot_mlf_core["project_slug"]}')
        try:
            repo.create_pull(title=pr_title, body=pr_body_text, head=f'mlf_core_sync_v{self.new_template_version}', base='development',
                             maintainer_can_modify=True)
            print('[bold blue]Successfully created PR!')

        # Something went wrong
        except GithubException as e:
            print(f'[bold red]Could not create a PR due to the following error:\n{e}')
            sys.exit(1)

    def check_pull_request_exists(self) -> bool:
        """
        Check, whether a mlf-core sync PR is already open. If so, close this one, as it is outdated.

        :return Whether a mlf-core sync PR is already open or not
        """
        repo = self.github.get_repo(f'{self.repo_owner}/{self.dot_mlf_core["project_slug"]}')
        # query all open PRs
        log.debug('Querying open PRs to check if a sync PR already exists.')
        # iterate over the open PRs of the repo to check if a mlf-core sync PR is open
        for pull_request in repo.get_pulls(state='open'):
            log.debug('Already open sync PR has been found.')
            # if an older, outdated mlf-core sync PR is still open, close it first
            if 'Important mlf-core template update' in pull_request.title:
                print('[bold blue]Detected open, but outdated mlf-core sync PR. Closing it in favor of the newest sync PR!')
                pull_request.edit(state='closed')
                return True
        return False

    def should_run_sync(self) -> bool:
        """
        Check, whether sync should run. This depends on two things:
        1.) First check, whether the user disabled sync in the mlf_core.cfg file with sync_enabled = False in the [sync] section
        2.) Secondly, whether the configured level in the [sync_level] section does not restrict the sync.
        Possible levels are:
            - patch: Always create a pull request (lower bound)
            - minor: Create a pull request if it's a minor or major change
            - major: Create a pull request only if it's a major change
        :return: Whether a sync should run determined by the rules above
        """
        log.debug(f'Checking sync level constraints using parsed results from {self.project_dir}/mlf_core.cfg')
        try:
            parser = ConfigParser()
            parser.read(f'{self.project_dir}/mlf_core.cfg')
            sync_enabled = parser.items('sync')
            # sync is enabled
            if sync_enabled[0][1].lower() in {'yes', 'y', 'true'}:
                pass
            # sync is disabled
            elif sync_enabled[0][1].lower() in {'no', 'n', 'false'}:
                return False
            # misconfigured sync_enabled with some unknown value
            else:
                print(f'[bold blue]Unknown value {sync_enabled[0][1]} for sync_enabled config.\nAllowed values are [bold green]Yes, yes, True, true, Y, y '
                      f'[bold blue] or [bold green]No, no, False, false, N, n [bold blue]!')
                return False
            # now, check for the sync level
            level_item = list(parser.items('sync_level'))
            log.debug(f'Parsing level constraint returned: {level_item}.')
            # check for proper configuration if the sync_level section (only one item named mlf_core_sync_level with valid levels major or minor
            if len(level_item) != 1 or 'sync_level' not in level_item[0][0] or not any(level_item[0][1] == valid_lvl for valid_lvl in
                                                                                       ['major', 'minor', 'patch']):
                print('[bold red]Your mlf_core_sync_level section is missconfigured. Make sure that it only contains one item named mlf_core_sync_level '
                      'with only valid levels patch, minor or major!')
                sys.exit(1)
            # check in case of minor update that level is not set to major (major case must not be handled as level is a lower bound)
            if self.patch_update:
                log.debug('Checking whether constraints allow patch updates.')
                return level_item[0][1] != 'minor' and level_item[0][1] != 'major'
            elif self.minor_update:
                log.debug('Checking whether constraints allow minor updates.')
                return level_item[0][1] != 'major'
            else:
                log.debug('All updates are allowed because patch level is set.')
                return True
        # mlf_core.cfg file was not found or has no section sync_level
        except NoSectionError:
            print('[bold red]Could not read from mlf_core.cfg file. '
                  'Make sure your specified path contains a mlf_core.cfg file and has a sync_level and sync section!')
            sys.exit(1)

    def get_blacklisted_sync_globs(self) -> list:
        """
        Get all blacklisted globs from the mlf_core.cfg file.
        :return: A list of all blacklisted globs for sync (file (types) that should not be included into the sync pull request)
        """
        try:
            parser = ConfigParser()
            parser.read(f'{self.project_dir}/mlf_core.cfg')
            globs = list(parser.items('sync_files_blacklisted'))
            nl = '\n'
            log.debug(f'Returning all blacklisted files globs parsed from {self.project_dir}/mlf_core.cfg.')
            log.debug(f'Blacklisted globs were {nl}{nl.join(glob[1] for glob in globs)}')
            return [glob[1] for glob in globs]

        # mlf_core.cfg file was not found or has no section called sync_files_blacklisted
        except NoSectionError:
            print('[bold red]Could not read from mlf_core.cfg file. Make sure your specified path contains a mlf_core.cfg file and has a '
                  'sync_files_blacklisted section!')
            sys.exit(1)

    def reset_target_dir(self):
        """
        Reset the target project directory. Check out the original branch.
        """
        print(f'[bold blue]Checking out original branch: {self.original_branch}')
        try:
            self.repo.git.checkout(self.original_branch)
        except git.exc.GitCommandError as e:
            print(f'[bold red]Could not reset to original branch {self.from_branch}:\n{e}')
            sys.exit(1)

    @staticmethod
    def update_sync_token(project_name: str, gh_username: str = '') -> None:
        """
        Update the sync token secret for the repository.

        :param project_name Name of the users project
        :param gh_username The Github username (only gets passed, if the repo is an orga repo)
        """
        gh_username = load_yaml_file(ConfigCommand.CONF_FILE_PATH)['github_username'] if not gh_username else gh_username
        # get the personal access token for user authentification
        log.debug('Asking for updated sync token value.')
        updated_sync_token = mlf_core_questionary_or_dot_mlf_core(function='password',
                                                                  question='Please enter your updated sync token value')
        print(f'[bold blue]\nUpdating sync secret for project {project_name}.')
        create_sync_secret(gh_username, project_name, updated_sync_token)
        print(f'[bold blue]\nSuccessfully updated sync secret for project {project_name}.')

    @staticmethod
    def has_template_version_changed(project_dir: Path) -> Tuple[bool, bool, bool, str, str]:
        """
        Check, if the mlf-core template has been updated since last check/sync of the user.

        :return: Both false if no versions changed or a micro change happened (for ex. 1.2.3 to 1.2.4). Return is_major_update True if a major version release
        happened for the mlf-core template (for example 1.2.3 to 2.0.0). Return is_minor_update True if a minor change happened (1.2.3 to 1.3.0).
        Return is_patch_update True if its a micro update (for example 1.2.3 to 1.2.4).
        mlf-core will use this to decide which syncing strategy to apply. Also return both versions.
        """
        # Try to compare against the development branch, since it is the most up to date (usually).
        # If a development branch does not exist compare against master.
        repo = git.Repo(project_dir)
        try:
            repo.git.checkout('development')
        except git.exc.GitCommandError:
            print('[bold red]Could not checkout development branch. Trying to checkout master...')
            try:
                repo.git.checkout('master')
            except git.exc.GitCommandError as e:
                print(f'[bold red]Could not checkout master branch.\n{e}')
                sys.exit(1)

        log.debug('Loading the project\'s template version and the mlf-core template version.')
        template_version_last_sync, template_handle = TemplateSync.sync_load_project_template_version_and_handle(project_dir)
        template_version_last_sync = version.parse(template_version_last_sync)  # type: ignore
        current_mlf_core_template_version = version.parse(TemplateSync.sync_load_template_version(template_handle))
        log.debug(f'Projects template version is {template_version_last_sync} and mlf-core template version is {current_mlf_core_template_version}')
        is_major_update, is_minor_update, is_patch_update = False, False, False

        # check if a major change happened (for example 1.2.3 to 2.0.0)
        if template_version_last_sync.major < current_mlf_core_template_version.major:  # type: ignore
            is_major_update = True
        # check if minor update happened (for example 1.2.3 to 1.3.0)
        elif template_version_last_sync.minor < current_mlf_core_template_version.minor:  # type: ignore
            is_minor_update = True
        # check if a patch update happened (for example 1.2.3 to 1.2.4)
        elif template_version_last_sync.micro < current_mlf_core_template_version.micro:  # type: ignore
            is_patch_update = True
        return is_major_update, is_minor_update, is_patch_update, str(template_version_last_sync), str(current_mlf_core_template_version)

    @staticmethod
    def sync_load_template_version(handle: str) -> str:
        """
        Load the version of the template available from mlf-core specified by the handler for syncing.

        :param handle: The template handle
        :return: The actual version number of the template in mlf-core
        """
        top_path = f'{os.path.dirname(__file__)}/..'
        available_templates_path = f'{str(top_path)}/create/templates/available_templates.yml'
        log.debug(f'Using available templates file from {available_templates_path} to load current mlf-core template version.')
        return load_mlf_core_template_version(handle, available_templates_path)

    @staticmethod
    def sync_load_project_template_version_and_handle(project_dir: Path) -> Tuple[str, str]:
        """
        Return the project template version since last sync for user (if no sync happened, return initial create version of the template)

        :param project_dir: Top level path to users project directory
        """
        return load_project_template_version_and_handle(project_dir)



import os
import sys
import logging
import appdirs
from pathlib import Path
from cryptography.fernet import Fernet
from ruamel.yaml import YAML
from rich import print
from rich.box import HEAVY_HEAD
from rich.style import Style
from rich.table import Table
from rich.console import Console

from mlf_core.common.levensthein_dist import most_similar_command
from mlf_core.custom_cli.questionary import mlf_core_questionary_or_dot_mlf_core
from mlf_core.common.load_yaml import load_yaml_file

log = logging.getLogger(__name__)


class ConfigCommand:
    """
    Class for the config command. Sets general configurations such as full name, email and Github username.
    """
    # path where the config file is stored for mlf_core
    CONF_FILE_PATH = f'{appdirs.user_config_dir(appname="mlf_core")}/mlf_core_cfg.yml'
    KEY_PAT_FILE = f'{appdirs.user_config_dir(appname="mlf_core")}/.ct_keys'

    @staticmethod
    def all_settings() -> None:
        """
        Set general settings and PAT.
        """
        if not ConfigCommand.check_mlf_core_config_dir_exists():
            ConfigCommand.create_config_dirs()
        ConfigCommand.config_general_settings()
        ConfigCommand.config_pat()

    @staticmethod
    def config_general_settings() -> None:
        """
        Set full_name, email and github username for reuse in any project created further on. Defaults for prompts
        are set to the previous configured values (if any).
        """
        already_configured = False
        settings = {}
        config_path = Path(ConfigCommand.CONF_FILE_PATH)
        if config_path.exists() and 'GITHUB_ACTIONS' not in os.environ:
            already_configured = True
            settings = load_yaml_file(ConfigCommand.CONF_FILE_PATH)

        full_name = mlf_core_questionary_or_dot_mlf_core(function='text',
                                                         question='Full name',
                                                         default='Homer Simpson' if not already_configured else settings['full_name'])
        email = mlf_core_questionary_or_dot_mlf_core(function='text',
                                                     question='Personal or work email',
                                                     default='homer.simpson@example.com' if not already_configured else settings['email'])
        github_username = mlf_core_questionary_or_dot_mlf_core(function='text',
                                                               question='Github username',
                                                               default='HomerGithub' if not already_configured else settings['github_username']).lower()

        # if the configs exist, just update them
        if os.path.exists(ConfigCommand.CONF_FILE_PATH):
            log.debug(f'Configuration was found at {ConfigCommand.CONF_FILE_PATH}. Updating configuration...')
            path = Path(ConfigCommand.CONF_FILE_PATH)
            yaml = YAML()
            settings = yaml.load(path)

            # update the full_name and email
            settings['full_name'] = full_name
            settings['email'] = email
            settings['github_username'] = github_username
            yaml.dump(settings, Path(ConfigCommand.CONF_FILE_PATH))

        # the configs don´t exist -> create them
        else:
            log.debug(f'Configuration was not found at {ConfigCommand.CONF_FILE_PATH}. Creating a new configuration file.')
            settings = {'full_name': full_name, 'email': email, 'github_username': github_username}
            yaml = YAML()
            yaml.dump(settings, Path(ConfigCommand.CONF_FILE_PATH))

    @staticmethod
    def config_pat() -> None:
        """
        Set the personal access token (PAT) for automatic Github repo creation.
        """
        if not ConfigCommand.check_mlf_core_config_dir_exists():
            ConfigCommand.create_config_dirs()

        try:
            path = Path(ConfigCommand.CONF_FILE_PATH)
            yaml = YAML()
            settings = yaml.load(path)
            if not all(attr in settings for attr in ['full_name', 'github_username', 'email']):
                print('[bold red]The mlf-core config file misses some required attributes!')
                print('[bold blue]Lets set them before setting your Github personal access token!')
                ConfigCommand.config_general_settings()

        except FileNotFoundError:
            print('[bold red]Cannot find a mlf-core config file. Is this your first time using mlf-core?')
            print('[bold blue]Lets create one before setting your Github personal access token!')
            ConfigCommand.config_general_settings()

        if mlf_core_questionary_or_dot_mlf_core(function='confirm',
                                                question='Do you want to configure your GitHub personal access token right now?\n'
                                                         'You can still configure it later '
                                                         'by calling    mlf-core config pat',
                                                default='Yes'):
            print('[bold blue]mlf-core requires your Github Access token to have full repository, workflow and create/update packages permissions!')
            access_token = mlf_core_questionary_or_dot_mlf_core(function='password',
                                                                question='Please enter your Github Access token')
            access_token_b = access_token.encode('utf-8')

            # ask for confirmation since this action will delete the PAT irrevocably if the user has not saved it anywhere else
            if not mlf_core_questionary_or_dot_mlf_core(function='confirm',
                                                        question='You´re about to update your personal access token. This action cannot be undone!\n'
                                                                 'Do you really want to continue?',
                                                        default='Yes'):
                sys.exit(1)

            # encrypt the given PAT and save the encryption key and encrypted PAT in separate files
            print('[bold blue]Generating key for encryption.')
            log.debug('Generating personal access key.')
            key = Fernet.generate_key()
            fer = Fernet(key)
            log.debug('Encrypting personal access token. ')
            print('[bold blue]Encrypting personal access token.')
            encrypted_pat = fer.encrypt(access_token_b)

            # write key
            with open(ConfigCommand.KEY_PAT_FILE, 'wb') as f:
                f.write(key)

            path = Path(ConfigCommand.CONF_FILE_PATH)
            yaml = YAML()
            settings = yaml.load(path)
            settings['pat'] = encrypted_pat
            log.debug(f'Dumping configuration to {ConfigCommand.CONF_FILE_PATH}')
            yaml.dump(settings, Path(ConfigCommand.CONF_FILE_PATH))

    @staticmethod
    def view_current_config() -> None:
        """
        Print the current users mlf-core configuration.
        """
        # load current settings
        try:
            log.debug(f'Fetching current cookietemple configuration at {ConfigCommand.CONF_FILE_PATH}.')
            settings = load_yaml_file(ConfigCommand.CONF_FILE_PATH)
            log.debug('Creating configuration table')
            # create the table and print
            table = Table(title="[bold]Your current configuration", title_style="blue", header_style=Style(color="blue", bold=True), box=HEAVY_HEAD)
            table.add_column('Name', style='green')
            table.add_column('Value', style='green')
            # add rows to the table consisting of the name and value of the current setting
            for (name, value) in settings.items():
                # don't print token directly, just inform it's set
                if name == 'pat':
                    table.add_row('[bold]Personal access token', 'TOKEN_IS_SET')
                else:
                    table.add_row(f'[bold]{name.capitalize().replace("_", " ")}', f'[white]{value}')
            # don't print PAT directly but inform if not set
            if 'pat' not in settings.keys():
                table.add_row('[bold]Personal access token', '[red]NO_TOKEN_SET')

            console = Console()
            console.print(table)

        except (FileNotFoundError, KeyError):
            print('[bold red]Did not found a mlf-core config file!\nIf this is your first time running mlf-core you can set them using mlf-core '
                  'config general')

    @staticmethod
    def similar_handle(section: str) -> None:
        """
        Try to use/suggest a similar handle if user misspelled it.

        :param section: The handle inputted by the user.
        """
        com_list, action = most_similar_command(section.lower(), {'general', 'pat', 'all'})
        # use best match
        if len(com_list) == 1 and action == 'use':
            print(f'[bold blue]Unknown handle {section}. Will use best match {com_list[0]}.\n')
            ConfigCommand.handle_switcher().get(com_list[0], lambda: 'Invalid handle!')()
        # suggest best match
        elif len(com_list) == 1 and action == 'suggest':
            print(f'[bold blue]Unknown handle {section}. Did you mean {com_list[0]}?')
            sys.exit(1)
            # multiple best matches
        elif len(com_list) > 1:
            nl = '\n'
            print(f'[bold red]Unknown handle \'{section}\'.\nMost similar handles are: [green]{nl}{nl.join(sorted(com_list))}')
        else:
            # unknown handle and no best match found
            print('[bold red]Unknown handle. [green]See mlf-core --help [red]for more information on valid handles')

    @staticmethod
    def check_mlf_core_config_dir_exists() -> bool:
        """
        Check whether the config directory for mlf-core exists.
        """
        log.debug(f'Checking whether a config directory already exists at {Path(ConfigCommand.CONF_FILE_PATH).parent}.')
        if not os.path.exists(Path(ConfigCommand.CONF_FILE_PATH).parent):
            log.debug('Config directory did not exist. Creating it.')
            return False
        return True

    @staticmethod
    def create_config_dirs() -> None:
        """
        Create the config directory, if none exists (and parent directories, if necessary)
        """
        os.makedirs(Path(ConfigCommand.CONF_FILE_PATH).parent)

    @staticmethod
    def handle_switcher() -> dict:
        """
        Just a helper to switch between handles.

        :return: The switcher with all handles.
        """
        switcher = {
            'all': ConfigCommand.all_settings,
            'general': ConfigCommand.config_general_settings,
            'pat': ConfigCommand.config_pat
        }
        return switcher



import os
from pathlib import Path
from dataclasses import dataclass

from mlf_core.create.github_support import prompt_github_repo
from mlf_core.create.template_creator import TemplateCreator
from mlf_core.create.domains.mlf_core_template_struct import MlfcoreTemplateStruct
from mlf_core.custom_cli.questionary import mlf_core_questionary_or_dot_mlf_core
from mlf_core.common.version import load_mlf_core_template_version


@dataclass
class TemplateStructMLflow(MlfcoreTemplateStruct):
    """
    MLFLOW-PYTORCH
    """
    command_line_interface: str = ''  # which command line library to use (click, argparse)
    testing_library: str = ''  # which testing library to use (pytest, unittest)


class MlflowCreator(TemplateCreator):

    def __init__(self):
        self.cli_struct = TemplateStructMLflow(domain='mlflow')
        super().__init__(self.cli_struct)
        self.WD_Path = Path(os.path.dirname(__file__))
        self.TEMPLATES_MLFLOW_PATH = f'{self.WD_Path.parent}/templates/mlflow'

        '"" TEMPLATE VERSIONS ""'
        self.MLFLOW_PYTORCH_TEMPLATE_VERSION = load_mlf_core_template_version('mlflow-pytorch', self.AVAILABLE_TEMPLATES_PATH)
        self.MLFLOW_TENSORFLOW_TEMPLATE_VERSION = load_mlf_core_template_version('mlflow-tensorflow', self.AVAILABLE_TEMPLATES_PATH)
        self.MLFLOW_XGBOOST_TEMPLATE_VERSION = load_mlf_core_template_version('mlflow-xgboost', self.AVAILABLE_TEMPLATES_PATH)

    def create_template(self, path: Path, dot_mlf_core: dict or None):
        """
        Handles the mlflow domain. Prompts the user for the language, general and domain specific options.
        """

        self.cli_struct.language = mlf_core_questionary_or_dot_mlf_core(function='select',
                                                                        question='Choose the project\'s primary framework',
                                                                        choices=['pytorch', 'tensorflow', 'xgboost'],
                                                                        default='pytorch',
                                                                        dot_mlf_core=dot_mlf_core,
                                                                        to_get_property='language')

        # prompt the user to fetch general template configurations
        super().prompt_general_template_configuration(dot_mlf_core)

        # switch case statement to prompt the user to fetch template specific configurations
        switcher = {
            'pytorch': self.mlflow_pytorch_options,
            'tensorflow': self.mlflow_tensorflow_options,
            'xgboost': self.mlflow_xgboost_options,
        }
        switcher.get(self.cli_struct.language)(dot_mlf_core)

        self.cli_struct.is_github_repo, \
            self.cli_struct.is_repo_private, \
            self.cli_struct.is_github_orga, \
            self.cli_struct.github_orga \
            = prompt_github_repo(dot_mlf_core)

        if self.cli_struct.is_github_orga:
            self.cli_struct.github_username = self.cli_struct.github_orga
        # create the chosen and configured template
        super().create_template_without_subdomain(self.TEMPLATES_MLFLOW_PATH)

        # switch case statement to fetch the template version
        switcher_version = {
            'pytorch': self.MLFLOW_PYTORCH_TEMPLATE_VERSION,
            'tensorflow': self.MLFLOW_TENSORFLOW_TEMPLATE_VERSION,
            'xgboost': self.MLFLOW_XGBOOST_TEMPLATE_VERSION,
        }
        self.cli_struct.template_version, self.cli_struct.template_handle = switcher_version.get(
            self.cli_struct.language), f'mlflow-{self.cli_struct.language.lower()}'

        # perform general operations like creating a GitHub repository and general linting
        super().process_common_operations(path=Path(path).resolve(), domain='mlflow', language=self.cli_struct.language, dot_mlf_core=dot_mlf_core)

    def mlflow_pytorch_options(self, dot_mlf_core: dict or None):
        """ Prompts for mlflow-pytorch specific options and saves them into the MlflowTemplateStruct """
        pass

    def mlflow_tensorflow_options(self, dot_mlf_core: dict or None):
        """ Prompts for mlflow-tensorflow specific options and saves them into the MlflowTemplateStruct """
        pass

    def mlflow_xgboost_options(self, dot_mlf_core: dict or None):
        """ Prompts for mlflow-xgboost specific options and saves them into the MlflowTemplateStruct """
        pass



import os

from mlf_core.common.load_yaml import load_yaml_file
from mlf_core.util.dict_util import is_nested_dictionary

AVAILABLE_TEMPLATES_PATH = f'{os.path.dirname(__file__)}/../create/templates/available_templates.yml'

# mlf-core main commands
MAIN_COMMANDS = ['create', 'lint', 'list', 'info', 'bump-version', 'sync', 'config', 'upgrade']
# the fraction relative to the commands length, a given input could differ from the real command to be automatically used instead
SIMILARITY_USE_FACTOR = 1 / 3
# the fraction relative to the commands length, a given input could differ from the real command to be suggested (if >1/3 of course)
SIMILARITY_SUGGEST_FACTOR = 2 / 3


def load_available_handles() -> set:
    """
    Load all available template handles.

    :return: A set of all available handles
    """
    available_templates = load_yaml_file(f'{AVAILABLE_TEMPLATES_PATH}')
    unsplit_handles = set()
    all_handles = set()
    nested_dict_to_handle_set(available_templates, unsplit_handles)
    all_handles.update(unsplit_handles)
    split_handles(unsplit_handles, all_handles)

    return all_handles


def nested_dict_to_handle_set(available_templates, unsplitted_handles: set) -> None:
    """
    Extract the handles from loaded yml file.

    :param available_templates: The loaded yml file as a (nested) dict
    :param unsplitted_handles: The set to save the handles
    """
    if is_nested_dictionary(available_templates):
        for templ in available_templates.values():
            if not is_nested_dictionary(templ):
                unsplitted_handles.add(templ['handle'])
            else:
                nested_dict_to_handle_set(templ, unsplitted_handles)
    else:
        # a single template to append was reached
        unsplitted_handles.add(available_templates['handle'])


def split_handles(unsplitted_handles, all_handles) -> None:
    """
    Split handles into all possible combinations.

    :param unsplitted_handles: A set of unsplitted handles
    :param all_handles: All handles mlf-core currently supports
    """
    for handle in unsplitted_handles:
        parts = handle.split('-')

        if len(parts) == 2:
            all_handles.add(parts[0])
        elif len(parts) == 3:
            all_handles.add(parts[0])
            all_handles.add(f'{parts[0]}-' + parts[1])



from dataclasses import dataclass


@dataclass
class MlfcoreTemplateStruct:
    """
    First section declares the variables that all template have in common
    """
    cookietemple_version: str = ''  # Version of mlf-core, which was used for creating the project
    domain: str = ''  # Domain of the template
    language: str = ''  # Primary language
    project_slug: str = ''  # Project name mlf-core uses for almost all further processing
    project_slug_no_hyphen: str = ''  # Project name mlf-core uses for almost all further processing without hyphens
    template_version: str = ''  # Version of the provided mlf-core template
    template_handle: str = ''  # Handle of the specific template, indicating which template is currently used
    github_username: str = ''  # Github username (in case of an organization repository, the organization name)
    creator_github_username: str = ''  # Github username of the person, that created the project
    is_github_repo: bool = False  # Whether the user wants to create a GitHub repo automatically
    is_repo_private: bool = False  # Whether to create a private Github repository
    is_github_orga: bool = False  # Whether Github repository is part of an organization
    github_orga: str = ''  # Name of the GitHub organization

    """
    This section contains some common attributes
    """
    full_name: str = ''  # Name of the template creator/organization
    email: str = ''  # Email of the creator
    project_name: str = ''  # Project's name the template is created for
    project_short_description: str = ''  # A short description of the project
    version: str = ''  # Version of the project; of the form: [0-9]*\.[0-9]*\.[0-9]*
    license: str = ''  # License of the project



import os
import sys
import logging
from collections import OrderedDict
import shutil
import re
import tempfile
import requests
from distutils.dir_util import copy_tree
from pathlib import Path
from dataclasses import asdict
from ruamel.yaml import YAML
from cookiecutter.main import cookiecutter

import mlf_core
from mlf_core.custom_cli.questionary import mlf_core_questionary_or_dot_mlf_core
from mlf_core.util.dir_util import delete_dir_tree
from mlf_core.create.github_support import create_push_github_repository, load_github_username, is_git_repo
from mlf_core.lint.lint import lint_project
from mlf_core.util.docs_util import fix_short_title_underline
from mlf_core.create.domains.mlf_core_template_struct import MlfcoreTemplateStruct
from mlf_core.config.config import ConfigCommand
from mlf_core.common.load_yaml import load_yaml_file
from rich import print

log = logging.getLogger(__name__)


class TemplateCreator:
    """
    The base class for all creators.
    It holds the basic template information that are common across all templates (like a project name).
    Furthermore it defines methods that are basic for the template creation process.
    """

    def __init__(self, creator_ctx: MlfcoreTemplateStruct):
        self.WD = os.path.dirname(__file__)
        self.TEMPLATES_PATH = f'{self.WD}/templates'
        self.COMMON_FILES_PATH = f'{self.TEMPLATES_PATH}/common_all_files'
        self.COMMON_MLFLOW_FILES_PATH = f'{self.TEMPLATES_PATH}/common_mlflow_files'
        self.AVAILABLE_TEMPLATES_PATH = f'{self.TEMPLATES_PATH}/available_templates.yml'
        self.AVAILABLE_TEMPLATES = load_yaml_file(self.AVAILABLE_TEMPLATES_PATH)
        self.CWD = Path.cwd()
        self.creator_ctx = creator_ctx

    def process_common_operations(self, path: Path, skip_fix_underline=False,
                                  domain: str = None, subdomain: str = None, language: str = None,
                                  dot_mlf_core: OrderedDict = None) -> None:
        """
        Create all stuff that is common for mlf-core's template creation process; in detail those things are:
        create and copy common files, fix docs style, lint the project and ask whether the user wants to create a github repo.
        """
        # create the common files and copy them into the templates directory
        self.create_common_files(domain='all', common_files_path=self.COMMON_FILES_PATH)
        # key in the switcher indicates, whether there are domain specific files or not (None)
        domain_switcher = {
            'mlflow': 'mlflow',
            'package': None
        }
        # if project is a project with domain specific files, copy all common files for the domain of this project
        try:
            domain_specific_files = domain_switcher[self.creator_ctx.domain]
        # this should only be the case, if mlf-core is developed further and new domains are added, therefore the error message
        except KeyError:
            print(f'[bold red]Unknown domain {self.creator_ctx.domain}! This domain seems to be new and must be added into the domain switcher!')
            sys.exit(1)
        if domain_specific_files:
            self.create_common_files(domain=domain_specific_files, common_files_path=self.COMMON_MLFLOW_FILES_PATH)

        self.create_dot_mlf_core(template_version=self.creator_ctx.template_version)

        project_path = f'{self.CWD}/{self.creator_ctx.project_slug}'

        # Ensure that docs are looking good (skip if flag is set)
        if not skip_fix_underline:
            fix_short_title_underline(f'{project_path}/docs/index.rst')

        # Lint the project to verify that the new template adheres to all standards
        lint_project(project_path)

        if self.creator_ctx.is_github_repo and not dot_mlf_core:
            # rename the currently created template to a temporary name, create Github repo, push, remove temporary template
            tmp_project_path = f'{project_path}_mlf_core_tmp'
            os.mkdir(tmp_project_path)
            create_push_github_repository(project_path, self.creator_ctx, tmp_project_path)
            shutil.rmtree(tmp_project_path, ignore_errors=True)

        if subdomain:
            print(f'\n[bold blue]Please visit: https://mlf-core.readthedocs.io/en/latest/available_templates/available_templates.html'
                  f'#{domain}-{subdomain}-{language} for more information about how to use your chosen template.')
        else:
            print(f'\n[bold blue]Please visit: https://mlf-core.readthedocs.io/en/latest/available_templates/available_templates.html'
                  f'#{domain}-{language} for more information about how to use your chosen template.')

        # do not move if path is current working directory or a directory named like the project in the current working directory (second is default case)
        if path != self.CWD and path != Path(self.CWD / self.creator_ctx.project_slug_no_hyphen):
            shutil.move(f'{self.CWD}/{self.creator_ctx.project_slug_no_hyphen}', f'{path}/{self.creator_ctx.project_slug_no_hyphen}')

    def create_template_without_subdomain(self, domain_path: str) -> None:
        """
        Creates a chosen template that does **not** have a subdomain.
        Calls cookiecutter on the main chosen template.

        :param domain_path: Path to the template, which is still in cookiecutter format
        """
        # Target directory is already occupied -> overwrite?
        occupied = os.path.isdir(f'{self.CWD}/{self.creator_ctx.project_slug}')
        if occupied:
            self.directory_exists_warning()

            # Confirm proceeding with overwriting existing directory
            if mlf_core_questionary_or_dot_mlf_core('confirm', 'Do you really want to continue?', default='Yes'):
                cookiecutter(f'{domain_path}/{self.creator_ctx.domain}_{self.creator_ctx.language.lower()}',
                             no_input=True,
                             overwrite_if_exists=True,
                             extra_context=self.creator_ctx_to_dict())
            else:
                print('[bold red]Aborted! Canceled template creation!')
                sys.exit(0)
        else:
            cookiecutter(f'{domain_path}/{self.creator_ctx.domain}_{self.creator_ctx.language.lower()}',
                         no_input=True,
                         overwrite_if_exists=True,
                         extra_context=self.creator_ctx_to_dict())

    def create_template_with_subdomain(self, domain_path: str, subdomain: str) -> None:
        """
        Creates a chosen template that **does** have a subdomain.
        Calls cookiecutter on the main chosen template.

        :param domain_path: Path to the template, which is still in cookiecutter format
        :param subdomain: Subdomain of the chosen template
        """
        occupied = os.path.isdir(f'{self.CWD}/{self.creator_ctx.project_slug}')
        if occupied:
            self.directory_exists_warning()

            # Confirm proceeding with overwriting existing directory
            if mlf_core_questionary_or_dot_mlf_core('confirm', 'Do you really want to continue?', default='Yes'):
                delete_dir_tree(Path(f'{self.CWD}/{self.creator_ctx.project_slug}'))
                cookiecutter(f'{domain_path}/{subdomain}_{self.creator_ctx.language.lower()}',
                             no_input=True,
                             overwrite_if_exists=True,
                             extra_context=self.creator_ctx_to_dict())

            else:
                print('[bold red]Aborted! Canceled template creation!')
                sys.exit(0)
        else:
            cookiecutter(f'{domain_path}/{subdomain}_{self.creator_ctx.language.lower()}',
                         no_input=True,
                         overwrite_if_exists=True,
                         extra_context=self.creator_ctx_to_dict())

    def create_template_with_subdomain_framework(self, domain_path: str, subdomain: str, framework: str) -> None:
        """
        Creates a chosen template that **does** have a subdomain.
        Calls cookiecutter on the main chosen template.

        :param domain_path: Path to the template, which is still in cookiecutter format
        :param subdomain: Subdomain of the chosen template
        :param framework: Chosen framework
        """
        occupied = os.path.isdir(f'{self.CWD}/{self.creator_ctx.project_slug}')
        if occupied:
            self.directory_exists_warning()

            # Confirm proceeding with overwriting existing directory
            if mlf_core_questionary_or_dot_mlf_core('confirm', 'Do you really want to continue?', default='Yes'):
                cookiecutter(f'{domain_path}/{subdomain}_{self.creator_ctx.language.lower()}/{framework}',
                             no_input=True,
                             overwrite_if_exists=True,
                             extra_context=self.creator_ctx_to_dict())

            else:
                print('[bold red]Aborted! Canceled template creation!')
                sys.exit(0)
        else:
            cookiecutter(f'{domain_path}/{subdomain}_{self.creator_ctx.language.lower()}/{framework}',
                         no_input=True,
                         overwrite_if_exists=True,
                         extra_context=self.creator_ctx_to_dict())

    def prompt_general_template_configuration(self, dot_mlf_core: OrderedDict):
        """
        Prompts the user for general options that are required by all templates.
        Options are saved in the creator context manager object.
        """
        try:
            if dot_mlf_core:
                # try to read name and email from existing config file
                self.creator_ctx.full_name = dot_mlf_core['full_name']
                self.creator_ctx.email = dot_mlf_core['email']
            else:
                self.creator_ctx.full_name = load_yaml_file(ConfigCommand.CONF_FILE_PATH)['full_name']
                self.creator_ctx.email = load_yaml_file(ConfigCommand.CONF_FILE_PATH)['email']
        except FileNotFoundError:
            # style and automatic use config
            print('[bold red]Cannot find a mlf_core config file. Is this your first time using mlf-core?')
            # inform the user and config all settings (with PAT optional)
            print('[bold blue]Lets set your name, email and Github username and you´re ready to go!')
            ConfigCommand.all_settings()
            # load mail and full name
            path = Path(ConfigCommand.CONF_FILE_PATH)
            yaml = YAML(typ='safe')
            settings = yaml.load(path)
            # set full name and mail
            if dot_mlf_core:
                self.creator_ctx.full_name = dot_mlf_core['full_name']
                self.creator_ctx.email = dot_mlf_core['email']
            else:
                self.creator_ctx.full_name = settings['full_name']
                self.creator_ctx.email = settings['email']

        self.creator_ctx.project_name = mlf_core_questionary_or_dot_mlf_core(function='text',
                                                                             question='Project name',
                                                                             default='Exploding Springfield',
                                                                             dot_mlf_core=dot_mlf_core,
                                                                             to_get_property='project_name').lower()

        # check if the project name is already taken on readthedocs.io
        # lower the string, since mlflow doesn't play with uppercase docker container names
        while self.readthedocs_slug_already_exists(self.creator_ctx.project_name) and not dot_mlf_core:
            print(f'[bold red]A project named {self.creator_ctx.project_name} already exists at readthedocs.io!')
            if mlf_core_questionary_or_dot_mlf_core(function='confirm',
                                                    question='Do you want to choose another name for your project?\n'
                                                             'Otherwise you will not be able to host your docs at readthedocs.io!', default='Yes'):
                self.creator_ctx.project_name = mlf_core_questionary_or_dot_mlf_core(function='text',
                                                                                     question='Project name',
                                                                                     default='Exploding Springfield').lower()
            # break if the project should be named anyways
            else:
                break
        self.creator_ctx.project_slug = self.creator_ctx.project_name.replace(' ', '_')
        self.creator_ctx.project_slug_no_hyphen = self.creator_ctx.project_slug.replace('-', '_')
        self.creator_ctx.project_short_description = mlf_core_questionary_or_dot_mlf_core(function='text',
                                                                                          question='Short description of your project',
                                                                                          default=f'{self.creator_ctx.project_name}'
                                                                                                  f'. A mlf-core based .',
                                                                                          dot_mlf_core=dot_mlf_core,
                                                                                          to_get_property='project_short_description')
        poss_vers = mlf_core_questionary_or_dot_mlf_core(function='text',
                                                         question='Initial version of your project',
                                                         default='0.1.0-SNAPSHOT',
                                                         dot_mlf_core=dot_mlf_core,
                                                         to_get_property='version')

        # make sure that the version has the right format
        while not re.match(r'(?<!.)\d+(?:\.\d+){2}(?:-SNAPSHOT)?(?!.)', poss_vers) and not dot_mlf_core:
            print('[bold red]The version number entered does not match semantic versioning.\n' +
                  'Please enter the version in the format \[number].\[number].\[number](-SNAPSHOT)!')  # noqa: W605
            poss_vers = mlf_core_questionary_or_dot_mlf_core(function='text',
                                                             question='Initial version of your project',
                                                             default='0.1.0-SNAPSHOT')
        self.creator_ctx.version = poss_vers

        self.creator_ctx.license = mlf_core_questionary_or_dot_mlf_core(function='select',
                                                                        question='License',
                                                                        choices=['MIT', 'BSD', 'ISC', 'Apache2.0', 'GNUv3', 'Boost', 'Affero',
                                                                                 'CC0', 'CCBY', 'CCBYSA', 'Eclipse', 'WTFPL', 'unlicence',
                                                                                 'Not open source'],
                                                                        default='MIT',
                                                                        dot_mlf_core=dot_mlf_core,
                                                                        to_get_property='license')

        if dot_mlf_core:
            self.creator_ctx.github_username = dot_mlf_core['github_username']
            self.creator_ctx.creator_github_username = dot_mlf_core['creator_github_username']
        else:
            self.creator_ctx.github_username = load_github_username()
            self.creator_ctx.creator_github_username = self.creator_ctx.github_username

    def create_common_files(self, domain: str, common_files_path: str) -> None:
        """
        Create a temporary directory for common files of a specified domain or all templates and apply cookiecutter on them.
        They are subsequently moved into the directory of the created template.
        """
        log.debug('Creating common files.')
        dirpath = tempfile.mkdtemp()
        copy_tree(common_files_path, dirpath)
        cwd_project = self.CWD
        os.chdir(dirpath)
        log.debug(f'Cookiecuttering common files at {dirpath}')
        cookiecutter(dirpath,
                     extra_context={'full_name': self.creator_ctx.full_name,
                                    'email': self.creator_ctx.email,
                                    'language': self.creator_ctx.language,
                                    'domain': self.creator_ctx.domain,
                                    'project_name': self.creator_ctx.project_name,
                                    'project_slug': self.creator_ctx.project_slug,
                                    'project_slug_no_hyphen': self.creator_ctx.project_slug_no_hyphen,
                                    'version': self.creator_ctx.version,
                                    'license': self.creator_ctx.license,
                                    'project_short_description': self.creator_ctx.project_short_description,
                                    'github_username': self.creator_ctx.github_username,
                                    'creator_github_username': self.creator_ctx.creator_github_username,
                                    'framework': self.creator_ctx.language.capitalize() if domain == 'mlflow' else ''},
                     no_input=True,
                     overwrite_if_exists=True)

        # copy the common files directory content to the created project
        log.debug('Copying common files into the created project')
        copy_tree(f'{Path.cwd()}/common_{domain}_files', f'{cwd_project}/{self.creator_ctx.project_slug}')
        # delete the tmp cookiecuttered common files directory
        log.debug('Delete common files directory.')
        delete_dir_tree(Path(f'{Path.cwd()}/common_{domain}_files'))
        shutil.rmtree(dirpath)
        # change to recent cwd so lint can run properly
        os.chdir(str(cwd_project))

    def readthedocs_slug_already_exists(self, project_name: str) -> bool:
        """
        Test whether there´s already a project with the same name on readthedocs
        :param project_name Name of the project the user wants to create
        """
        print(f'[bold blue]Looking up {project_name} at readthedocs.io!')
        try:
            request = requests.get(f'https://{project_name.replace(" ", "")}.readthedocs.io')
            if request.status_code == 200:
                return True
        # catch exceptions when server may be unavailable or the request timed out
        except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
            print('[bold red]Cannot check whether name already taken on readthedocs.io because its unreachable at the moment!')
            return False

    def directory_exists_warning(self) -> None:
        """
        If the directory is already a git directory within the same project, print error message and exit.
        Otherwise print a warning that a directory already exists and any further action on the directory will overwrite its contents.
        """
        if is_git_repo(Path(f'{self.CWD}/{self.creator_ctx.project_slug}')):
            print(f'[bold red]Error: A git project named {self.creator_ctx.project_slug} already exists at [green]{self.CWD}\n')
            print('[bold red]Aborting!')
            sys.exit(1)
        else:
            print(f'[bold yellow]WARNING: [red]A directory named {self.creator_ctx.project_slug} already exists at [blue]{self.CWD}\n')
            print('Proceeding now will overwrite this directory and its content!')

    def create_dot_mlf_core(self, template_version: str):
        """
        Overrides the version with the version of the template.
        Dumps the configuration for the template generation into a .mlf_core yaml file.

        :param template_version: Version of the specific template
        """
        log.debug('Creating .mlf_core.yml file.')
        self.creator_ctx.template_version = f'{template_version} # <<MLF-CORE_NO_BUMP>>'
        self.creator_ctx.mlf_core_version = f'{mlf_core.__version__} # <<MLF-CORE_NO_BUMP>>'
        with open(f'{self.creator_ctx.project_slug}/.mlf_core.yml', 'w') as f:
            yaml = YAML()
            struct_to_dict = self.creator_ctx_to_dict()
            yaml.dump(struct_to_dict, f)

    def creator_ctx_to_dict(self) -> dict:
        """
        Create a dict from the our Template Structure dataclass
        :return: The dict containing all key-value pairs with non empty values
        """
        return {key: val for key, val in asdict(self.creator_ctx).items() if val != ''}



import sys
import os
from rich.console import Console


def force_terminal_in_github_action() -> Console:
    """
    Check, whether the GITHUB_ACTIONS environment variable is set or not.
    If set, the process runs in a workflow file and we need to tell rich, in order to get colored output as well.
    """
    if 'GITHUB_ACTIONS' in os.environ:
        return Console(file=sys.stderr, force_terminal=True)
    else:
        return Console(file=sys.stderr)


# the console used for printing with rich
console = force_terminal_in_github_action()



#!/usr/bin/env python

"""The setup script."""

import os
import mlf_core as module
from setuptools import setup, find_packages


def walker(base, *paths):
    file_list = set([])
    cur_dir = os.path.abspath(os.curdir)

    os.chdir(base)
    try:
        for path in paths:
            for dname, dirs, files in os.walk(path):
                for f in files:
                    file_list.add(os.path.join(dname, f))
    finally:
        os.chdir(cur_dir)

    return list(file_list)


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('CHANGELOG.rst') as history_file:
    history = history_file.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3', ]

setup(
    author="Lukas Heumos",
    author_email='lukas.heumos@posteo.net',
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="Reproducible machine learning pipelines using mlflow.",
    entry_points={
        'console_scripts': [
            'mlf-core=mlf_core.cli:main',
        ],
    },
    install_requires=requirements,
    license="Apache2.0",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='mlf-core',
    name='mlf-core',
    packages=find_packages(include=['mlf_core', 'mlf_core.*']),
    package_data={
        module.__name__: walker(
            os.path.dirname(module.__file__),
            'create/templates'
        ),
    },
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/mlf-core/mlf-core',
    version='1.11.0',
    zip_safe=False,
)
